{
TGMElevation component

  ES: componente para calcular elevaciones de terreno de un conjunto de LatLng
  EN: component for calculate a terrain elevations of a set of LatLng

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap, establecer
      las lat/lng y ejecutar
  EN: put the component into a form, link to a TGMMap, establish the lat/lng and
      execute
=========================================================================
History:

ver 0.1.8:
  ES- primera versin
  EN- first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMElevation;

interface

uses
  Classes, Contnrs, DB,
  GMMap, GMLatLng, GMClasses, GMConstants, GMMarker, GMPolyline;

type
  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#ElevationResult
  **************************************************************************** }
  TElevationResult = class
  private
    FLocation: TLatLng;
    FResolution: Real;
    FElevation: Real;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    property Location: TLatLng read FLocation;
    property Elevation: Real read FElevation;
    property Resolution: Real read FResolution;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TElevationResults = class
  private
    FElevationResult: TObjectList;
    FStatus: TElevationStatus;
    function GetItems(Index: Integer): TElevationResult;
    function GetCount: Integer;
  protected
    function AddElevationResult(Elevation, Resolution, Lat, Lng: string): Integer;
    procedure ParseElevations(Elevations: string);
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Clear;

    property Status: TElevationStatus read FStatus;
    property Items[Index: Integer]: TElevationResult read GetItems; default;
    property Count: Integer read GetCount;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#ElevationService
    https://developers.google.com/maps/documentation/javascript/elevation
    https://developers.google.com/maps/documentation/elevation/index
  **************************************************************************** }
  TGMElevation = class(TGMObjects, ILinePoint)
  private
    FLinePoints: TLinePoints;
    FSamples: Integer;
    FElevationType: TElevationType;
    FElevationResult: TElevationResults;
    function GetCountLinePoints: Integer;
    function GetItems(I: Integer): TLinePoint;
    procedure SetSamples(const Value: Integer);
  protected
    procedure DeleteMapObjects; override;
    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;

    { ILinePoint }
    procedure LinePointChanged;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    function AddLatLng(LatLng: TLatLng): TLinePoint; overload;
    function AddLatLng(Lat, Lng: Real): TLinePoint; overload;
    procedure AddLatLngFromCSV(LatColumn, LngColumn: Integer; FileName: string;
      Delimiter: Char = ','; DeleteBeforeLoad: Boolean = True; WithRownTitle: Boolean = True);
    procedure AddLatLngFromDataSet(DataSet: TDataSet; LatField, LngField: string;
      DeleteBeforeLoad: Boolean = True);
    procedure AddLatLngFromBasePolyline(Poly: TBasePolyline; DeleteBeforeLoad: Boolean = True);
    procedure Clear;
    procedure DelLatLng(Index: Integer);

    procedure Execute;

    property CountLinePoints: Integer read GetCountLinePoints;
    property Items[I: Integer]: TLinePoint read GetItems; default;
    property ElevationResult: TElevationResults read FElevationResult;
  published
    property LinePoints: TLinePoints read FLinePoints write FLinePoints;
    property Samples: Integer read FSamples write SetSamples;
    property ElevationType: TElevationType read FElevationType write FElevationType;
  end;

implementation

uses
  SysUtils,
  Lang, GMFunctions, WebControl, GMPolygon;

{ TGMElevation }

function TGMElevation.AddLatLng(LatLng: TLatLng): TLinePoint;
begin
  Result := AddLatLng(LatLng.Lat, LatLng.Lng);
end;

function TGMElevation.AddLatLng(Lat, Lng: Real): TLinePoint;
begin
  Result := FLinePoints.Add;
  Result.Lat := Lat;
  Result.Lng := Lng;
end;

procedure TGMElevation.AddLatLngFromCSV(LatColumn, LngColumn: Integer;
  FileName: string; Delimiter: Char; DeleteBeforeLoad, WithRownTitle: Boolean);
var
  L1, L2: TStringList;
  i: Integer;
  Point: TLinePoint;
  Stop: Boolean;
begin
  if not FileExists(FileName) then
    raise Exception.Create(GetTranslateText('El fichero no existe', Language));

  if DeleteBeforeLoad then FLinePoints.Clear;

  L1 := TStringList.Create;
  L2 := TStringList.Create;
  try
    L1.LoadFromFile(FileName);
    if WithRownTitle then L1.Delete(0);

    L2.Delimiter := Delimiter;
    {$IF CompilerVersion > 16}
    L2.StrictDelimiter := True;
    {$IFEND}

    for i := 0 to L1.Count - 1 do
    begin
      L2.DelimitedText := L1[i];

      if (LatColumn > L2.Count) or (LngColumn > L2.Count) then
        raise Exception.Create(GetTranslateText('El nmero de columna es incorrecto', Language));

      Point := LinePoints.Add;
      Point.Lat := Point.StringToReal(L2[LatColumn]);
      Point.Lng := Point.StringToReal(L2[LngColumn]);

      Stop := False;
      if Stop then Break;
    end;
  finally
    if Assigned(L1) then FreeAndNil(L1);
    if Assigned(L2) then FreeAndNil(L2);
  end;
end;

procedure TGMElevation.AddLatLngFromDataSet(DataSet: TDataSet; LatField,
  LngField: string; DeleteBeforeLoad: Boolean);
var
  Bkm: TBookmark;
  Point: TLinePoint;
begin
  if not DataSet.Active then DataSet.Open;

  if DeleteBeforeLoad then FLinePoints.Clear;

  DataSet.DisableControls;
  try
    Bkm := DataSet.GetBookmark;
    DataSet.First;
    while not DataSet.Eof do
    begin
      Point := LinePoints.Add;
      Point.Lat := DataSet.FieldByName(LatField).AsFloat;
      Point.Lng := DataSet.FieldByName(LngField).AsFloat;

      DataSet.Next;
    end;
  finally
    DataSet.GotoBookmark(Bkm);
    DataSet.FreeBookmark(Bkm);
    DataSet.EnableControls;
  end;
end;

procedure TGMElevation.AddLatLngFromBasePolyline(Poly: TBasePolyline;
  DeleteBeforeLoad: Boolean);
var
  Point: TLinePoint;
  i: Integer;
begin
  if DeleteBeforeLoad then FLinePoints.Clear;

  for i := 0 to Poly.CountLinePoints - 1 do
  begin
    Point := LinePoints.Add;
    Point.Assign(Poly[i]);
  end;

  if (Poly is TPolygon) and (Poly.CountLinePoints > 0) then
  begin
    Point := LinePoints.Add;
    Point.Assign(Poly[0]);
  end;
end;

procedure TGMElevation.Clear;
begin
  FLinePoints.Clear;
  FElevationResult.Clear;
end;

constructor TGMElevation.Create(aOwner: TComponent);
begin
  inherited;

  FLinePoints := TLinePoints.Create(Self, TLinePoint);
  FSamples := 2;
  FElevationType := etForLocations;
  FElevationResult := TElevationResults.Create;
end;

procedure TGMElevation.DeleteMapObjects;
begin
  inherited;

end;

procedure TGMElevation.DelLatLng(Index: Integer);
begin
  FLinePoints.Delete(Index);
end;

destructor TGMElevation.Destroy;
begin
  if Assigned(FLinePoints) then FreeAndNil(FLinePoints);
  if Assigned(FElevationResult) then FreeAndNil(FElevationResult);

  inherited;
end;

procedure TGMElevation.EventFired(EventType: TEventType;
  Params: array of const);
begin
  inherited;

end;

procedure TGMElevation.Execute;
const
  StrParams = '%s,%s,%s';
var
  Params: string;
  WC: TWebControl;
  Tmp: string;
begin
  if not Assigned(Map) then
    raise Exception.Create(GetTranslateText('Mapa no asignado', Language));

  Params := Format(StrParams, [
                  QuotedStr(TTransform.ElevationTypeToStr(FElevationType)),
                  QuotedStr(LinePoints.PointsToStr(Map.Precision)),
                  IntToStr(Samples)
                  ]);
  ExecuteScript('GetElevations', Params);

  ElevationResult.Clear;
  WC := TWebControl.Create(Map.WebBrowser);
  try
    Tmp := WC.GetStringField(ElevationsForm, ElevationsFormStatus);
    ElevationResult.FStatus := TTransform.StrToElevationStatus('es' + Tmp);
    Tmp := WC.GetStringField(ElevationsForm, ElevationsFormElevations);
    ElevationResult.ParseElevations(Tmp);
  finally
    FreeAndNil(WC);
  end;
end;

function TGMElevation.GetCountLinePoints: Integer;
begin
  Result := LinePoints.Count;
end;

function TGMElevation.GetItems(I: Integer): TLinePoint;
begin
  Result := TLinePoint(FLinePoints[i]);
end;

procedure TGMElevation.LinePointChanged;
begin

end;

procedure TGMElevation.SetSamples(const Value: Integer);
begin
  if FSamples = Value then Exit;

  FSamples := Value;
  if FSamples < 2 then FSamples := 2;
  if FSamples > 512 then FSamples := 512;
end;

procedure TGMElevation.ShowElements;
begin
  inherited;

end;

{ TElevationResults }

function TElevationResults.AddElevationResult(Elevation, Resolution, Lat,
  Lng: string): Integer;
var
  ElevRes: TElevationResult;
begin
  ElevRes := TElevationResult.Create;
  ElevRes.FLocation.Lat := ElevRes.FLocation.StringToReal(Lat);
  ElevRes.FLocation.Lng := ElevRes.FLocation.StringToReal(Lng);
  ElevRes.FResolution := ElevRes.FLocation.StringToReal(Resolution);
  ElevRes.FElevation := ElevRes.FLocation.StringToReal(Elevation);
  Result := FElevationResult.Add(ElevRes);
end;

procedure TElevationResults.Clear;
begin
  FStatus := esNO_REQUEST;
  FElevationResult.Clear;
end;

constructor TElevationResults.Create;
begin
  FElevationResult := TObjectList.Create;
end;

destructor TElevationResults.Destroy;
begin
  if Assigned(FElevationResult) then FreeAndNil(FElevationResult);

  inherited;
end;

function TElevationResults.GetCount: Integer;
begin
  Result := FElevationResult.Count;
end;

function TElevationResults.GetItems(Index: Integer): TElevationResult;
begin
  Result := TElevationResult(FElevationResult[Index]);
end;

procedure TElevationResults.ParseElevations(Elevations: string);
var
  T1, T2: TStringList;
  i: Integer;
begin
  T1 := TStringList.Create;
  T2 := TStringList.Create;
  try
    T1.Delimiter := ';';
    T2.Delimiter := '|';
    {$IF CompilerVersion > 16}
    T1.StrictDelimiter := True;
    T2.StrictDelimiter := True;
    {$IFEND}

    T1.DelimitedText := Elevations;
    for i := 0 to T1.Count - 1 do
    begin
      T2.DelimitedText := T1[i];
      if T2.Count <> 4 then Continue;

      AddElevationResult(T2[0], T2[1], T2[2], T2[3]);
    end;
  finally
    if Assigned(T1) then FreeAndNil(T1);
    if Assigned(T2) then FreeAndNil(T2);
  end;
end;

{ TElevationResult }

constructor TElevationResult.Create;
begin
  FLocation := TLatLng.Create;
  FResolution := 0;
  FElevation := 0;
end;

destructor TElevationResult.Destroy;
begin
  if Assigned(FLocation) then FreeAndNil(FLocation);

  inherited;
end;

end.
